@@ -42,139 +42,93 @@
 struct vrg_priv {
 	unsigned		magic;
 #define VRG_PRIV_MAGIC		0xb886e711
+	struct req		*req;
 	ssize_t			range_low;
 	ssize_t			range_high;
 	ssize_t			range_off;
 };
 
 static int v_matchproto_(vdp_fini_f)
-vrg_range_fini(struct req *req, void **priv)
+vrg_range_fini(struct vdp_ctx *vdc, void **priv)
 {
 	struct vrg_priv *vrg_priv;
 
-	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-	CAST_OBJ_NOTNULL(vrg_priv, *priv, VRG_PRIV_MAGIC);
-	if (vrg_priv->range_off < vrg_priv->range_high)
-		Req_Fail(req, SC_RANGE_SHORT);
-	*priv = NULL;	/* struct on ws, no need to free */
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	TAKE_OBJ_NOTNULL(vrg_priv, priv, VRG_PRIV_MAGIC);
+	if (vrg_priv->req->resp_len >= 0 &&
+	    vrg_priv->range_off < vrg_priv->range_high) {
+		Req_Fail(vrg_priv->req, SC_RANGE_SHORT);
+		vrg_priv->req->vdc->retval = -1;
+	}
+	/* struct on ws, no need to free */
 	return (0);
 }
 
 static int v_matchproto_(vdp_bytes_f)
-vrg_range_bytes(struct req *req, enum vdp_action act, void **priv,
+vrg_range_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
     const void *ptr, ssize_t len)
 {
 	int retval = 0;
-	ssize_t l;
+	ssize_t l = 0;
 	const char *p = ptr;
 	struct vrg_priv *vrg_priv;
 
-	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
 	AN(priv);
 	CAST_OBJ_NOTNULL(vrg_priv, *priv, VRG_PRIV_MAGIC);
 
-	l = vrg_priv->range_low - vrg_priv->range_off;
-	if (l > 0) {
-		if (l > len)
-			l = len;
-		vrg_priv->range_off += l;
-		p += l;
-		len -= l;
+	if (ptr != NULL) {
+		l = vrg_priv->range_low - vrg_priv->range_off;
+		if (l > 0) {
+			if (l > len)
+				l = len;
+			vrg_priv->range_off += l;
+			p += l;
+			len -= l;
+		}
+		l = vmin(vrg_priv->range_high - vrg_priv->range_off, len);
+		vrg_priv->range_off += len;
+		if (vrg_priv->range_off >= vrg_priv->range_high)
+			act = VDP_END;
 	}
-	l = vrg_priv->range_high - vrg_priv->range_off;
-	if (l > len)
-		l = len;
+
 	if (l > 0)
-		retval = VDP_bytes(req, act, p, l);
-	else if (act > VDP_NULL)
-		retval = VDP_bytes(req, act, p, 0);
-	vrg_priv->range_off += len;
-	return (retval ||
-	    vrg_priv->range_off >= vrg_priv->range_high ? 1 : 0);
+		retval = VDP_bytes(vdc, act, p, l);
+	else if (l == 0 && act > VDP_NULL)
+		retval = VDP_bytes(vdc, act, p, 0);
+	return (retval || act == VDP_END ? 1 : 0);
 }
 
 /*--------------------------------------------------------------------*/
 
 static const char *
-vrg_dorange(struct req *req, const char *r, void **priv)
+vrg_dorange(struct req *req, void **priv)
 {
-	ssize_t low, high, has_low, has_high, t;
+	ssize_t low, high;
 	struct vrg_priv *vrg_priv;
+	const char *err;
 
-	if (strncasecmp(r, "bytes=", 6))
-		return ("Not Bytes");
-	r += 6;
+	err = http_GetRange(req->http, &low, &high, req->resp_len);
+	if (err != NULL)
+		return (err);
 
-	/* The low end of range */
-	has_low = low = 0;
-	while (vct_isdigit(*r)) {
-		has_low = 1;
-		t = low;
-		low *= 10;
-		low += *r++ - '0';
-		if (low < t)
-			return ("Low number too big");
-	}
+	if (low < 0 || high < 0)
+		return (NULL);		// Allow 200 response
 
-	if (*r++ != '-')
-		return ("Missing hyphen");
-
-	/* The high end of range */
-	has_high = high = 0;
-	while (vct_isdigit(*r)) {
-		has_high = 1;
-		t = high;
-		high *= 10;
-		high += *r++ - '0';
-		if (high < t)
-			return ("High number too big");
-	}
-
-	if (*r != '\0')
-		return ("Trailing stuff");
-
-	if (has_high + has_low == 0)
-		return ("Neither high nor low");
-
-	if (!has_low) {
-		if (req->resp_len < 0)
-			return (NULL);		// Allow 200 response
-		if (high == 0)
-			return ("No low, high is zero");
-		low = req->resp_len - high;
-		if (low < 0)
-			low = 0;
-		high = req->resp_len - 1;
-	} else if (req->resp_len >= 0 && (high >= req->resp_len || !has_high))
-		high = req->resp_len - 1;
-	else if (!has_high || req->resp_len < 0)
-		return (NULL);			// Allow 200 response
-	/*
-	 * else (bo != NULL) {
-	 *    We assume that the client knows what it's doing and trust
-	 *    that both low and high make sense.
-	 * }
-	 */
-
-	if (high < low)
-		return ("high smaller than low");
-
-	if (req->resp_len >= 0 && low >= req->resp_len)
-		return ("low range beyond object");
-
-	if (req->resp_len >= 0)
+	if (req->resp_len >= 0) {
 		http_PrintfHeader(req->resp, "Content-Range: bytes %jd-%jd/%jd",
 		    (intmax_t)low, (intmax_t)high, (intmax_t)req->resp_len);
-	else
+		req->resp_len = (intmax_t)(1 + high - low);
+	} else
 		http_PrintfHeader(req->resp, "Content-Range: bytes %jd-%jd/*",
 		    (intmax_t)low, (intmax_t)high);
-	req->resp_len = (intmax_t)(1 + high - low);
 
 	vrg_priv = WS_Alloc(req->ws, sizeof *vrg_priv);
 	if (vrg_priv == NULL)
 		return ("WS too small");
 
 	INIT_OBJ(vrg_priv, VRG_PRIV_MAGIC);
+	vrg_priv->req = req;
 	vrg_priv->range_off = 0;
 	vrg_priv->range_low = low;
 	vrg_priv->range_high = high + 1;
@@ -206,6 +160,7 @@ vrg_ifrange(struct req *req)
 			return (0);
 		if ((e[0] == 'W' && e[1] == '/'))	// rfc7232,l,547,548
 			return (0);
+		/* XXX: should we also have http_etag_cmp() ? */
 		return (strcmp(p, e) == 0);		// rfc7232,l,548,548
 	}
 
@@ -215,7 +170,7 @@ vrg_ifrange(struct req *req)
 		return (0);
 
 	/* the response needs a Date */
-	// rfc7232 fc7232,l,439,440
+	// rfc7232,l,439,440
 	if (!http_GetHdr(req->resp, H_Date, &p))
 		return (0);
 	d = VTIM_parse(p);
@@ -241,29 +196,40 @@ vrg_ifrange(struct req *req)
 }
 
 static int v_matchproto_(vdp_init_f)
-vrg_range_init(struct req *req, void **priv)
+vrg_range_init(VRT_CTX, struct vdp_ctx *vdc, void **priv)
 {
-	const char *r;
 	const char *err;
 
-	assert(http_GetHdr(req->http, H_Range, &r));
-	if (!vrg_ifrange(req))	// rfc7233,l,455,456
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_ORNULL(ctx->req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	AN(priv);
+
+	if (ctx->req == NULL) {
+		VSLb(vdc->vsl, SLT_Error,
+		     "range can only be used on the client side");
 		return (1);
-	err = vrg_dorange(req, r, priv);
+	}
+
+	// not using vdc->{hd,cl}, because range needs req anyway for Req_Fail()
+
+	if (!vrg_ifrange(ctx->req))		// rfc7233,l,455,456
+		return (1);
+	err = vrg_dorange(ctx->req, priv);
 	if (err == NULL)
 		return (*priv == NULL ? 1 : 0);
 
-	VSLb(req->vsl, SLT_Debug, "RANGE_FAIL %s", err);
-	if (req->resp_len >= 0)
-		http_PrintfHeader(req->resp,
+	VSLb(vdc->vsl, SLT_Debug, "RANGE_FAIL %s", err);
+	if (ctx->req->resp_len >= 0)
+		http_PrintfHeader(ctx->req->resp,
 		    "Content-Range: bytes */%jd",
-		    (intmax_t)req->resp_len);
-	http_PutResponse(req->resp, "HTTP/1.1", 416, NULL);
+		    (intmax_t)ctx->req->resp_len);
+	http_PutResponse(ctx->req->resp, "HTTP/1.1", 416, NULL);
 	/*
 	 * XXX: We ought to produce a body explaining things.
 	 * XXX: That really calls for us to hit vcl_synth{}
 	 */
-	req->resp_len = 0;
+	ctx->req->resp_len = 0;
 	return (1);
 }
 
@@ -273,3 +239,83 @@ const struct vdp VDP_range = {
 	.bytes =	vrg_range_bytes,
 	.fini =		vrg_range_fini,
 };
+
+/*--------------------------------------------------------------------*/
+
+int
+VRG_CheckBo(struct busyobj *bo)
+{
+	ssize_t rlo, rhi, crlo, crhi, crlen, clen;
+	const char *err;
+
+	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+
+	if (!cache_param->http_range_support)
+		return (0);
+
+	err = http_GetRange(bo->bereq0, &rlo, &rhi, -1);
+	clen = http_GetContentLength(bo->beresp);
+	crlen = http_GetContentRange(bo->beresp, &crlo, &crhi);
+
+	if (err != NULL) {
+		VSLb(bo->vsl, SLT_Error, "Invalid range header (%s)", err);
+		return (-1);
+	}
+
+	if (crlen < -1) {
+		VSLb(bo->vsl, SLT_Error, "Invalid content-range header");
+		return (-1);
+	}
+
+	if (clen < -1) {
+		VSLb(bo->vsl, SLT_Error, "Invalid content-length header");
+		return (-1);
+	}
+
+	if (crlo < 0 && crhi < 0 && crlen < 0) {
+		VSLb(bo->vsl, SLT_Debug,
+		    "Missing content-range header or unknown range unit");
+		return (0);
+	}
+
+	if (rlo < 0 && rhi < 0) {
+		VSLb(bo->vsl, SLT_Error, "Unexpected content-range header");
+		return (-1);
+	}
+
+	if (crlo < 0) {		// Content-Range: bytes */123
+		assert(crhi < 0);
+		assert(crlen > 0);
+		if (http_GetStatus(bo->beresp) == 416)
+			return (0);
+		crlo = 0;
+		crhi = crlen - 1;
+	}
+
+#define RANGE_CHECK(val, op, crval, what)			\
+	do {							\
+		if (val >= 0 && !(val op crval)) {		\
+			VSLb(bo->vsl, SLT_Error,		\
+			    "Expected " what " %zd, got %zd",	\
+			    crval, val);			\
+			return (-1);				\
+		}						\
+	} while (0)
+
+	crlen = (crhi - crlo) + 1;
+	RANGE_CHECK(clen, ==, crlen, "content length");
+
+	/* NB: if the client didn't specify a low range the high range
+	 * was adjusted based on the resource length, and a high range
+	 * is allowed to be out of bounds so at this point there is
+	 * nothing left to check.
+	 */
+	if (rlo < 0)
+		return (0);
+
+	RANGE_CHECK(rlo, ==, crlo, "low range");
+	RANGE_CHECK(rhi, >=, crhi, "minimum high range");
+#undef RANGE_CHECK
+
+	return (0);
+}
