@@ -169,7 +169,7 @@ http_write(const struct http *hp, int lvl,
 }
 
 static int
-get_bytes(const struct http *hp, char *buf, int n)
+get_bytes(const struct http *hp, char *buf, size_t n)
 {
 	int i;
 	struct pollfd pfd[1];
@@ -197,7 +197,7 @@ get_bytes(const struct http *hp, char *buf, int n)
 		i = read(hp->sess->fd, buf, n);
 		if (!(pfd[0].revents & POLLIN))
 			vtc_log(hp->vl, 4,
-			    "HTTP2 rx poll (fd:%d revents: %x n=%d, i=%d)",
+			    "HTTP2 rx poll (fd:%d revents: %x n=%zu, i=%d)",
 			    hp->sess->fd, pfd[0].revents, n, i);
 		if (i == 0)
 			vtc_log(hp->vl, 3,
@@ -467,22 +467,20 @@ decode_hdr(struct http *hp, struct hpk_hdr *h, const struct vsb *vsb)
 		r = HPK_DecHdr(iter, h + n);
 		if (r == hpk_err )
 			break;
-		vtc_log(hp->vl, 4,
-				"header[%2d]: %s : %s",
-				n,
-				h[n].key.ptr,
-				h[n].value.ptr);
+		vtc_log(hp->vl, 4, "header[%2d]: %s: %s",
+		    n, h[n].key.ptr, h[n].value.ptr);
 		n++;
 		if (r == hpk_done)
 			break;
 	}
 
-	if (r != hpk_done)
+	if (r != hpk_done) {
 		vtc_log(hp->vl, hp->fatal ? 4 : 0,
-				"Header decoding failed (%d) %d", r, hp->fatal);
-	else if (n == MAX_HDR)
+		    "Header decoding failed (%d) %d", r, hp->fatal);
+	} else if (n == MAX_HDR) {
 		vtc_log(hp->vl, hp->fatal,
-				"Max number of headers reached (%d)", MAX_HDR);
+		    "Max number of headers reached (%d)", MAX_HDR);
+	}
 
 	HPK_FreeIter(iter);
 }
@@ -804,7 +802,7 @@ receive_frame(void *priv)
 			vtc_fatal(s->vl, "Expected CONTINUATION frame for "
 			    "stream %u", expect_cont);
 
-		/* parse the frame according to it type, and fill the metada */
+		/* parse the frame according to it type, and fill the metadata */
 		switch (f->type) {
 			case TYPE_DATA:
 				parse_data(s, f);
@@ -1011,7 +1009,7 @@ cmd_var_resolve(const struct stream *s, const char *spec, char *buf)
 	/* SECTION: stream.spec.zexpect.settings SETTINGS specific
 	 *
 	 * settings.ack
-	 *	"true" if the ACK flag was set, else ""false.
+	 *	"true" if the ACK flag was set, else "false".
 	 *
 	 * settings.push
 	 *	"true" if the push settings was set to yes, "false" if set to
@@ -1900,6 +1898,9 @@ cmd_txprio(CMD_ARGS)
  * \-hdrsize INT
  *	maximum size of the header list authorized
  *
+ * \-0xHH[HH] INT
+ *	tx arbitraty settings with tag xx
+ *
  * \-ack
  *	set the ack bit
  */
@@ -1908,7 +1909,8 @@ cmd_txsettings(CMD_ARGS)
 {
 	struct stream *s, *s2;
 	struct http *hp;
-	char *p;
+	char *p, *e;
+	unsigned long u;
 	uint32_t val = 0;
 	struct frame f;
 	//TODO dynamic alloc
@@ -1954,6 +1956,15 @@ cmd_txsettings(CMD_ARGS)
 			PUT_KV(av, vl, framesize, val, 0x5);
 		else if (!strcmp(*av, "-hdrsize"))
 			PUT_KV(av, vl, hdrsize, val, 0x6);
+		else if (!strncmp(*av, "-0x", 3)) {
+			p = *av + 3;
+			errno = 0;
+			u = strtoul(p, &e, 16);
+			if (*p == '\0' || *e != '\0' || u > 0xffff || errno != 0)
+				vtc_fatal(vl, "Invalid settings tag %s", p);
+			assert(u <= 0xffff);
+			PUT_KV(av, vl, hdrtbl, val, (uint16_t)u);
+		}
 		else if (!strcmp(*av, "-ack"))
 			f.flags |= 1;
 		else
@@ -2881,10 +2892,16 @@ b64_settings(const struct http *hp, const char *s)
 			buf = "unknown";
 
 		if (v == 1) {
-			if (hp->sfd)
-				assert(HPK_ResizeTbl(hp->encctx, v) != hpk_err);
-			else
-				assert(HPK_ResizeTbl(hp->decctx, v) != hpk_err);
+			enum hpk_result hrs;
+			if (hp->sfd) {
+				AN(hp->encctx);
+				hrs = HPK_ResizeTbl(hp->encctx, v);
+			} else {
+				AN(hp->decctx);
+				hrs = HPK_ResizeTbl(hp->decctx, v);
+			}
+			if (hrs != hpk_done)
+				vtc_fatal(hp->vl, "HPK resize failed %d\n", hrs);
 		}
 
 		vtc_log(hp->vl, 4, "Upgrade: %s (%d): %ju",
