@@ -243,8 +243,20 @@ int __llvm_gcov_flush(void);
 
 #if __GNUC_PREREQ__(4, 3) || defined(__clang__)
 #  define v_cold_    __attribute__((cold))
+#  define v_used_(x) __attribute__((x, used))
 #else
 #  define v_cold_
+#  define v_used_(x)
+#endif
+
+#if defined __has_attribute
+#  if __has_attribute(counted_by)
+#    define v_counted_by_(field) __attribute__((counted_by(field)))
+#  endif
+#endif
+
+#ifndef v_counted_by_
+#  define v_counted_by_(field)
 #endif
 
 /* VTIM API overhaul WIP */
@@ -263,6 +275,9 @@ typedef struct {
 
 #define Tcheck(t)	do { (void)pdiff((t).b, (t).e); } while (0)
 #define Tlen(t)		(pdiff((t).b, (t).e))
+#define Tstr(s)		(/*lint -e(446)*/ (txt){(s), (s) + strlen(s)})
+#define Tstrcmp(t, s)	(strncmp((t).b, (s), Tlen(t)))
+#define Tforeach(c, t)	for ((c) = (t).b; (c) < (t).e; (c)++)
 
 /* #3020 dummy definitions until PR is merged*/
 #define LIKELY(x)	(x)
