@@ -81,6 +81,13 @@ vcc_utype(vcc_type_t t)
 	return (t->name);
 }
 
+static vcc_type_t
+vcc_stringstype(vcc_type_t t)
+{
+	return (t->stringform ? STRINGS : t);
+}
+
+
 static void vcc_expr0(struct vcc *tl, struct expr **e, vcc_type_t fmt);
 static void vcc_expr_cor(struct vcc *tl, struct expr **e, vcc_type_t fmt);
 static void vcc_expr_typecheck(struct vcc *tl, struct expr **e, vcc_type_t fmt,
@@ -152,13 +159,18 @@ static void
 vcc_strands_edit(const struct expr *e1, const struct expr *e2)
 {
 
-	if (e2->nstr == 1) {
-		VSB_printf(e1->vsb, "TOSTRAND(%s)", VSB_data(e2->vsb));
-		return;
-	}
+	assert(e2->fmt == STRANDS || e2->fmt == STRINGS);
 
-	VSB_printf(e1->vsb, "TOSTRANDS(%d,\v+\n%s\v-)",
-	    e2->nstr, VSB_data(e2->vsb));
+	if (e2->fmt == STRANDS)
+		VSB_cat(e1->vsb, VSB_data(e2->vsb));
+	else if (e2->nstr == 0)
+		VSB_printf(e1->vsb, "vrt_null_strands");
+	else if (e2->nstr == 1)
+		VSB_printf(e1->vsb, "TOSTRAND(%s)", VSB_data(e2->vsb));
+	else {
+		VSB_printf(e1->vsb, "TOSTRANDS(%d,\v+\n%s\v-)",
+		   e2->nstr, VSB_data(e2->vsb));
+	}
 }
 
 static struct expr *
@@ -288,19 +300,18 @@ vcc_expr_tobool(struct vcc *tl, struct expr **e)
  */
 
 static void
-vcc_expr_tostring(struct vcc *tl, struct expr **e, vcc_type_t fmt)
+vcc_expr_tostring(struct vcc *tl, struct expr **e)
 {
 	const char *p;
 	uint8_t	constant = EXPR_VAR;
 
 	CHECK_OBJ_NOTNULL(*e, EXPR_MAGIC);
-	assert(fmt == STRINGS || fmt->stringform);
-	assert(fmt != (*e)->fmt);
+	assert((*e)->fmt != STRINGS);
 
 	p = (*e)->fmt->tostring;
 	if (p != NULL) {
 		AN(*p);
-		*e = vcc_expr_edit(tl, fmt, p, *e, NULL);
+		*e = vcc_expr_edit(tl, STRINGS, p, *e, NULL);
 		(*e)->constant = constant;
 		(*e)->nstr = 1;
 	} else {
@@ -442,20 +453,20 @@ struct func_arg {
 	VTAILQ_ENTRY(func_arg)	list;
 };
 
-static void
-vcc_do_enum(struct vcc *tl, struct func_arg *fa, int len, const char *ptr)
+static struct expr *
+vcc_do_enum(struct vcc *tl, const char *cfunc, int len, const char *ptr)
 {
 	const char *r;
 
 	(void)tl;
-	r = strchr(fa->cname, '.');
+	r = strchr(cfunc, '.');
 	AN(r);
-	fa->result = vcc_mk_expr(VOID, "*%.*s.enum_%.*s",
-	    (int)(r - fa->cname), fa->cname, len, ptr);
+	return (vcc_mk_expr(VOID, "*%.*s.enum_%.*s",
+	    (int)(r - cfunc), cfunc, len, ptr));
 }
 
 static void
-vcc_do_arg(struct vcc *tl, struct func_arg *fa)
+vcc_do_arg(struct vcc *tl, const char *cfunc, struct func_arg *fa)
 {
 	struct expr *e2;
 	struct vjsn_val *vv;
@@ -474,7 +485,7 @@ vcc_do_arg(struct vcc *tl, struct func_arg *fa)
 			vcc_ErrWhere(tl, tl->t);
 			return;
 		}
-		vcc_do_enum(tl, fa, PF(tl->t));
+		fa->result = vcc_do_enum(tl, cfunc, PF(tl->t));
 		SkipToken(tl, ID);
 	} else {
 		if (fa->type == SUB)
@@ -540,7 +551,6 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 		assert(vjsn_is_array(vv));
 		fa = calloc(1, sizeof *fa);
 		AN(fa);
-		fa->cname = cfunc;
 		VTAILQ_INSERT_TAIL(&head, fa, list);
 
 		vvp = VTAILQ_FIRST(&vv->children);
@@ -548,7 +558,7 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 			fa->result = vcc_priv_arg(tl, vvp->value, sym);
 			vvp = VTAILQ_NEXT(vvp, list);
 			if (vvp != NULL)
-				fa->name = vvp->value;
+				fa->cname = fa->name = vvp->value;
 			continue;
 		}
 		fa->type = VCC_Type(vvp->value);
@@ -557,6 +567,9 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 		if (vvp != NULL) {
 			fa->name = vvp->value;
 			vvp = VTAILQ_NEXT(vvp, list);
+			AN(vvp); /* vmod_syntax 2.0 */
+			fa->cname = vvp->value;
+			vvp = VTAILQ_NEXT(vvp, list);
 			if (vvp != NULL) {
 				fa->val = vvp->value;
 				vvp = VTAILQ_NEXT(vvp, list);
@@ -583,7 +596,7 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 			if (t1->tok == '=')
 				break;
 		}
-		vcc_do_arg(tl, fa);
+		vcc_do_arg(tl, cfunc, fa);
 		if (tl->err)
 			VSB_printf(tl->sb, "Expected argument: %s %s\n\n",
 			    fa->type->name,
@@ -615,7 +628,7 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 		}
 		vcc_NextToken(tl);
 		SkipToken(tl, '=');
-		vcc_do_arg(tl, fa);
+		vcc_do_arg(tl, cfunc, fa);
 		ERRCHK(tl);
 		if (tl->t->tok == ')')
 			break;
@@ -632,18 +645,18 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 	VTAILQ_FOREACH_SAFE(fa, &head, list, fa2) {
 		n++;
 		if (fa->optional) {
-			AN(fa->name);
+			AN(fa->cname);
 			bprintf(ssa, "\v1.valid_%s = %d,\n",
-			    fa->name, fa->avail);
+			    fa->cname, fa->avail);
 			e1 = vcc_expr_edit(tl, e1->fmt, ssa, e1, NULL);
 		}
 		if (fa->result == NULL && fa->type == ENUM && fa->val != NULL)
-			vcc_do_enum(tl, fa, strlen(fa->val), fa->val);
+			fa->result = vcc_do_enum(tl, cfunc, strlen(fa->val), fa->val);
 		if (fa->result == NULL && fa->val != NULL)
 			fa->result = vcc_mk_expr(fa->type, "%s", fa->val);
 		if (fa->result != NULL && sa != NULL) {
-			if (fa->name)
-				bprintf(ssa, "\v1.%s = \v2,\n", fa->name);
+			if (fa->cname)
+				bprintf(ssa, "\v1.%s = \v2,\n", fa->cname);
 			else
 				bprintf(ssa, "\v1.arg%d = \v2,\n", n);
 			e1 = vcc_expr_edit(tl, e1->fmt, ssa, e1, fa->result);
@@ -816,13 +829,13 @@ vcc_expr5(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 			ERRCHK(tl);
 			/* Unless asked for a HEADER, fold to string here */
 			if (*e && fmt != HEADER && (*e)->fmt == HEADER) {
-				vcc_expr_tostring(tl, e, STRINGS);
+				vcc_expr_tostring(tl, e);
 				ERRCHK(tl);
 			}
 			return;
 		}
 		VSB_printf(tl->sb,
-		    "Symbol '%.*s' type (%s) can not be used in expression.\n",
+		    "Symbol '%.*s' type (%s) cannot be used in expression.\n",
 		    PF(t), sym->kind->name);
 		vcc_ErrWhere(tl, t);
 		if (sym->def_b != NULL) {
@@ -1071,9 +1084,9 @@ vcc_expr_add(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 		} else if (tk->tok == '+' &&
 		    ((*e)->fmt == STRINGS || fmt == STRINGS)) {
 			if ((*e)->fmt != STRINGS)
-				vcc_expr_tostring(tl, e, STRINGS);
+				vcc_expr_tostring(tl, e);
 			if (e2->fmt != STRINGS)
-				vcc_expr_tostring(tl, &e2, STRINGS);
+				vcc_expr_tostring(tl, &e2);
 			if (vcc_islit(*e) && vcc_isconst(e2)) {
 				lit = vcc_islit(e2);
 				*e = vcc_expr_edit(tl, STRINGS,
@@ -1188,7 +1201,7 @@ cmp_string(struct vcc *tl, struct expr **e, const struct cmps *cp)
 	vcc_NextToken(tl);
 	vcc_expr_add(tl, &e2, STRINGS);
 	ERRCHK(tl);
-	if (e2->fmt != STRINGS) {
+	if (vcc_stringstype(e2->fmt) != STRINGS) {
 		VSB_printf(tl->sb,
 		    "Comparison of different types: %s '%.*s' %s\n",
 		    vcc_utype((*e)->fmt), PF(tk), vcc_utype(e2->fmt));
@@ -1262,11 +1275,13 @@ vcc_expr_cmp(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 	tk = tl->t;
 
 	for (cp = vcc_cmps; cp->fmt != VOID; cp++) {
-		if ((*e)->fmt == cp->fmt && tl->t->tok == cp->token) {
-			AN(cp->func);
-			cp->func(tl, e, cp);
-			return;
-		}
+		if (tl->t->tok != cp->token)
+			continue;
+		if (vcc_stringstype((*e)->fmt) != cp->fmt)
+			continue;
+		AN(cp->func);
+		cp->func(tl, e, cp);
+		return;
 	}
 
 	switch (tk->tok) {
@@ -1343,7 +1358,7 @@ vcc_expr_bin_bool(struct vcc *tl, struct expr **e, vcc_type_t fmt,
 	ERRCHK(tl);
 	if ((*e)->fmt != BOOL) {
 		VSB_printf(tl->sb,
-		    "'%s' must be preceeded by BOOL,"
+		    "'%s' must be preceded by BOOL,"
 		    " found %s.\n", tokstr, vcc_utype((*e)->fmt));
 		vcc_ErrWhere2(tl, tk, tl->t);
 		return;
@@ -1420,7 +1435,7 @@ vcc_expr0(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 		return;
 
 	if ((*e)->fmt != STRINGS && fmt->stringform)
-		vcc_expr_tostring(tl, e, STRINGS);
+		vcc_expr_tostring(tl, e);
 
 	if ((*e)->fmt->stringform) {
 		VSB_printf(tl->sb, "Cannot convert type %s(%s) to %s(%s)\n",
@@ -1431,7 +1446,7 @@ vcc_expr0(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 	}
 
 	if (fmt == BODY && !(*e)->fmt->bodyform)
-		vcc_expr_tostring(tl, e, STRINGS);
+		vcc_expr_tostring(tl, e);
 
 	if (fmt == BODY && (*e)->fmt->bodyform) {
 		if ((*e)->fmt == STRINGS)
@@ -1593,14 +1608,13 @@ vcc_Eval_Default(struct vcc *tl, struct expr **e, struct token *t,
 	(void)sym;
 	(void)t;
 
-	if (fmt == PROBE)
-		*e = vcc_mk_expr(PROBE, "%s", vcc_default_probe(tl));
-	else if (fmt == BACKEND)
-		*e = vcc_mk_expr(BACKEND, "*(VCL_conf.default_director)");
-	else {
+	if (fmt->default_sym == NULL) {
 		VSB_cat(tl->sb, "Symbol 'default' is a reserved word.\n");
 		vcc_ErrWhere(tl, t);
+		return;
 	}
+
+	*e = vcc_mk_expr(fmt, "%s", fmt->default_sym->rname);
 }
 
 /*--------------------------------------------------------------------
@@ -1637,6 +1651,6 @@ vcc_Expr_Init(struct vcc *tl)
 
 	sym = VCC_MkSym(tl, "default", SYM_MAIN, SYM_FUNC, VCL_LOW, VCL_HIGH);
 	AN(sym);
-	sym->type = BACKEND;	// ... can also (sometimes) deliver PROBE
+	sym->type = DEFAULT;
 	sym->eval = vcc_Eval_Default;
 }
